config = {
    "host": "https://127.0.0.1:2746",
    "token": None,
    "k8s_config_file": None,
    "k8s_api_server": None,
    "private_key_host_path": "/home/docker/.ssh",
    "save_path_as_parameter": False,
    "catalog_dir_name": ".dflow",
    "archive_mode": "tar",
    "util_image": "python:3.8",
    "util_image_pull_policy": None,
    "extender_image": "dptechnology/dflow-extender",
    "extender_image_pull_policy": None,
    "dispatcher_image": "dptechnology/dpdispatcher",
    "dispatcher_image_pull_policy": None,
    "save_keys_in_global_outputs": True,
    "mode": "default",
}
